# THIS FILE BELONGS TO THE METATYPE1 PACKAGE
#
# It is an AWK script that does the main job of convertion from
# METAPOST output to (raw) Type 1 format
# History:
# 29.04.2003: ver. 0.35, a bug in gawk circumvented (in clear_hints)
# 18.01.2003: ver. 0.34, normal round function introduced; old one
#             renamed to fround (forced round); no more mess due
#             ADL comment in PFB and Ascender and Descender ones in AFM;
#             again fun. rational: superfluously robust ;-)
# 06.01.2003: ver. 0.33, fun. rational: unlikely case included (as a comment)
# 03.01.2003: ver. 0.32, fontdimens and headerbytes handled
# 31.07.2002: banner added
# 14.07.2002: comment (question) added in `pickup_stem'
# 22.09.2001: ver. 0.31 (remarks moved to an informal readme.his file)
# 26.01.2001: an empty set of KPX pairs admitted
# 14.01.2001: functions rational and approx added

BEGIN {
# HEAD
# CONSOLE="/dev/tty"
# for MS-DOS you may use:
  CONSOLE="CON"
  mess("This is mp2pf, ver. 0.34 (a converter from MP EPSes to PS Type 1 font)")
  if (NAME=="") {mess("!NAME MUST NOT BE EMPTY"); exit(1);}
  if (LOG=="") LOG=NAME ".clg" # converter log
  if (AFM=="") AFM=NAME ".afm"
  if (PFB=="") PFB=NAME ".p"
  if (FD=="") FD=NAME ".pfi"
  if (KD=="") KD=NAME ".kpx"
  if (CD=="") CD="pfcommon.dat"
  if (PL=="") PL="piclist.tex"
  max_font_dimen=99; max_header_byte=99;
  mllx=10000; mlly=10000; murx=-10000; mury=-10000
  get_data_files()
  # dominant stems chosen manually have the highest priority:
  if ("STDVW" in font_data) vstem_stat[font_data["STDVW"]]=10000
  if ("STDHW" in font_data) hstem_stat[font_data["STDHW"]]=10000
  subrs_base=4; subrs[""]=-1 # standard empty subr has number 3==subrs_base-1
  # the following code must be consistent with fontbase.mp and mpform.sty
  while (getline < PL >0) if (/^\\EPSNAMEandNUMBER{.*}{[0-9]+}/) {
    sub(/^\\EPSNAMEandNUMBER{.*}{/,""); sub(/} *$/,"")
    file[++file[0]]=NAME "." $0
  }
# BODY
  for (curr_file=1; curr_file<=file[0]; curr_file++)
    while (getline < file[curr_file] >0) {
      if (/^%%BoundingBox:/) {
        if (++n%4==0) printf "#" > CONSOLE
        clear_hints(2)
        name=""; code=0; hsb=width=0; x0=y0=0; #sentinels
        curr_path=acc_path=start_acc=0; hints_pos=0;
        hsb=llx=$2; lly=$3; urx=$4; ury=$5
        if (mllx>llx) mllx=llx; if (mlly>lly) mlly=lly
        if (murx<urx) murx=urx; if (mury<ury) mury=ury
      }
      if (/^%GLYNFO:/) {
        if ($2=="NAME") {name=$3; code=$4}
        if ($2=="HSBW") {#hsb=$3; # obsolete
          width=$4; # if (int(width)!=width) width=round(width*1000)/1000
          x0=hsb; y0=0} # hsbw should precede hints
        if ($2=="ACC") {acc_path=$3}
        if ($2=="HHINT") {pickup_stem(hstem,hcov,hstem_stat,$3,$4,20.5)}
        if ($2=="VHINT") {pickup_stem(vstem,vcov,vstem_stat,$3-hsb,$4,0)}
        if ($2=="BEGINCHAR") {
          find_stem3(hstem,hstem3,20.5); find_stem3(vstem,vstem3,0)
          store_pfb("/" name " {\n\t" fround(hsb) " " approx(width) " hsbw")
          hints0_pos=mark_pfb_pos()
          store_afm("C " code " ; WX " width " ; N " name\
            " ; B " llx " " lly " " urx " " ury " ;", code, name)
        }
      }
      if (/lineto/) {
        set_hints($1,$2)
        make_lineto($1-x0,$2-y0)
        x0=$1; y0=$2
      } else
      if (/curveto/) {
        set_hints($5,$6)
        make_curveto($1-x0,$2-y0,$3-$1,$4-$2,$5-$3,$6-$4)
        x0=$5; y0=$6
      } else
      if (/moveto/) {
        # $1=="newpath"
        if (++curr_path==acc_path) {# we treat accent nearly as a new glyph
          change_hints(); pfb_text[hints0_pos]=flush_hints(hints0)
          clear_hints(1); hints_pos=mark_pfb_pos(); hints0_pos=mark_pfb_pos()
          start_acc=1
        }
        set_hints($2,$3)
        make_moveto($2-x0,$3-y0)
        x0=$2; y0=$3; x_path0=x0; y_path0=y0
      }
      if (/closepath/) {
        set_hints(x_path0,y_path0)
        store_pfb("\tclosepath")
      }
      if (/showpage/) {
        change_hints()
        pfb_text[hints0_pos]=flush_hints(hints0)
        store_pfb("\tendchar\n\t} ND")
      }
    } # end while, for
# TAIL
  print "" > CONSOLE
  stdhw=make_stem_stat(hstem_stat,stemsnaph)
  stdvw=make_stem_stat(vstem_stat,stemsnapv)
  flush_afm(); flush_pfb()
}

# FUNCTIONS

function make_stem_stat(stat,snap, m,d,i,j,t) {
  while (snap[0]<12) { # 12 is max length of stemsnap array
    m=0; d=0
    for (i in stat) if (stat[i]>m) {m=stat[i]; d=i+0}
    if (m>1) # ignore single stems
      {snap[++snap[0]]=d; delete stat[d]} else break
  }
  d=snap[1]
  for (i=1; i<snap[0]; ++i) {
    m=i; for (j=i+1; j<=snap[0]; j++) if (snap[j]<snap[m]) m=j
    if (m!=i) {t=snap[i]; snap[i]=snap[m]; snap[m]=t}
  }
  return d
}

function clear_hints(lev, i) {
  # lev==0 at hints replacement (clears curr stems)
  hints[0]=0
  for (i in curr_hstem) delete curr_hstem[i]
  for (i in curr_vstem) delete curr_vstem[i]
  if (lev>0) hints0[0]=0 # lev==1 at the begin of accent
  if (lev>1) {# lev==2 at the begin of a glyph (clears all stem info)
    # `i=i' is a circumvention of a gawk bug (improper typing
    # scalar vs. array)
    for (i in curr0_hstem) {i=i; delete curr0_hstem[i]}
    for (i in curr0_vstem) {i=i; delete curr0_vstem[i]}
    for (i in hstem) {i=i; delete hstem[i]}
    for (i in vstem) {i=i; delete vstem[i]}
    for (i in hstem3) {i=i; delete hstem3[i]}
    for (i in vstem3) {i=i; delete vstem3[i]}
    for (i in hcov) {i=i; delete hcov[i]}
    for (i in vcov) {i=i; delete vcov[i]}
  }
}

function pickup_stem(stem,covering,stem_stat,b,d,ghost, i,a,b_,d_) {
  b+=0; d+=0 # must be numbers; should be integers: to check or not to check?
  for (i in stem) {
    split(stem[i],a); b_=a[1]+0; d_=a[2]+0 # must also be a number
    if ((b_+d_>=b) && (b_<=b+d)) if ((b_!=b) || (d_!=d)) {
      covering[b_,d_,b,d]; covering[b,d,b_,d_]
    }
  }
# We remember only the ``smallest'' hint for a given point (both
# starting and ending positions are taken into account); because we
# are not going to minimize the number of hint replacements, the complete
# list of feasible hints for a given point is not needed. As a side-effect,
# we obtain the following convenient behaviour: assume that, e.g., in
# dieresis the distance between dots is equal to a dot diameter; METAPOST
# likely will generate three vstems -- two for the dots and one for the gap;
# the latter will be removed by `pickup_stem'.
  fix_min_stem(stem,b,b,d); fix_min_stem(stem,b+d,b,d)
  if (abs(d-ghost)>1) stem_stat[d]++
}

function fix_min_stem(stem,b0,b,d, a) {
  if (stem[b0]=="") stem[b0]= b " " d
  else {split(stem[b0],a); if (a[2]+0>d) stem[b0]= b " " d}
}

function find_stem3(stem,stem3,ghost, i,j,a,b1,b2,b3,d1,d2,d3,gap,found) {
  for (i in stem) {
    for (j in stem) {
      if (i!=j) {
          split(stem[i],a); b1=a[1]+0; d1=a[2]+0 # must be numbers
          split(stem[j],a); b2=a[1]+0; d2=a[2]+0 # must be numbers
          gap=b2-(b1+d1)
          if ((gap>0) && (abs(d1-ghost)>1)) {# no ``ghost'' triple stems
            b3=b2+d2+gap; d3=d1
            if ((b3 in stem ? stem[b3]==(b3 " " d3) : 0) ||
               ((b3+d3) in stem ? stem[b3+d3]==(b3 " " d3) : 0)) {
                 stem3[b1]=stem3[b1+d1]=stem3[b2]=stem3[b2+d2]=\
                   stem3[b3]=stem3[b3+d3]= b1 " " d1 " " d2 " " gap
                 found=1
            }
          }
        }
      if (found) break
    }
    if (found) break
  }
  if (found) for (i in stem) delete stem[i] # Adobe documentation demands it
}

function set_hints(x,y, do_rep) {
  x-=hsb; y+=0 # must be numbers
  if (x in vstem3) set_hint3(x, vstem3, curr0_vstem, hints0, "v")
  else if (x in vstem ? !hint_clash(x, vstem, curr0_vstem, vcov) : 0)
    set_hint(x, vstem, curr0_vstem, hints0, "v")
  if (y in hstem3) set_hint3(y, hstem3, curr0_hstem, hints0, "h")
  else if (y in hstem ? !hint_clash(y, hstem, curr0_hstem, hcov) : 0)
    set_hint(y, hstem, curr0_hstem, hints0, "h")
#
  if (x in vstem) do_rep+=hint_clash(x, vstem, curr_vstem, vcov)
  if (y in hstem) do_rep+=hint_clash(y, hstem, curr_hstem, hcov)
  if (do_rep) {change_hints(); clear_hints(0); hints_pos=mark_pfb_pos()}
#
  if (x in vstem3) set_hint3(x, vstem3, curr_vstem, hints, "v")
  else if (x in vstem) set_hint(x, vstem, curr_vstem, hints, "v")
  if (y in hstem3) set_hint3(y, hstem3, curr_hstem, hints, "h")
  else if (y in hstem) set_hint(y, hstem, curr_hstem, hints, "h")
}

function hint_clash(p,stem,curr_stem,cov,  a,i,clash) {
  if (!(p in curr_stem)) {
    split(stem[p],a)
    for (i in curr_stem) if (curr_stem[i]>0)
      if ((i,curr_stem[i],a[1],a[2]) in cov) {clash=1; break}
  }
  return clash
}

function set_hint3(p,stem3,curr_stem,hints,pfx,  a) {
  if (!(p in curr_stem)) {
    split(stem3[p],a)
    store_hint(hints,sprintf("%1s %4d %4d %4d %4d",pfx,a[1],a[2],a[3],a[4]))
    curr_stem[a[1]]=curr_stem[a[1]+a[2]+a[3]+2*a[4]]=a[2]
    curr_stem[a[1]+a[2]]=curr_stem[a[1]+2*a[2]+a[3]+2*a[4]]=-a[2]
    curr_stem[a[1]+a[2]+a[4]]=a[3]; curr_stem[a[1]+a[2]+a[3]+a[4]]=-a[3]
  }
}

function set_hint(p,stem,curr_stem,hints,pfx,  a) {
  if (!(p in curr_stem)) {
    split(stem[p],a)
    store_hint(hints,sprintf("%1s %4d %4d",pfx,a[1],a[2]))
    curr_stem[a[1]]=a[2]; curr_stem[a[1]+a[2]]=-a[2]
  }
}

function store_hint(hints,hint_txt, i,t) {
  hints[++hints[0]]=hint_txt
  for (i=hints[0]; i>1; --i) if (hints[i]>hints[i-1]) {# trivial sort
    t=hints[i]; hints[i]=hints[i-1]; hints[i-1]=t
  } else break
}

function flush_hints(hints, i,a,r) {
  for (i=1; i<=hints[0]; ++i)
    if (i in hints) # in accent some hints are removed
      if (split(hints[i],a)>3)
        r=r (r==""?"":"\n") "\t" a[2] " " a[3] " " a[2]+a[3]+a[5] " " a[4] \
          " " a[2]+a[3]+a[4]+2*a[5] " " a[3] " " a[1] "stem3"
      else
        r=r (r==""?"":"\n") "\t" a[2] " " a[3] " " a[1] "stem"
  return r
}

function change_hints( i,j) {
  if (start_acc==1) {
    start_acc=0
    i=j=1; while ((i<=hints[0]) && (j<=hints0[0])) {# don't put hints twice
      if (hints[i]==hints0[j]) {delete hints[i]; j++}; i++
    }
  }
  if (hints_pos!=0) {
    i=flush_hints(hints)
    if (!(i in subrs)) {xsubrs[++xsubrs[0]]=i; subrs[i]=xsubrs[0]}
    pfb_text[hints_pos]="\t" subrs[i]+subrs_base " 4 callsubr"
      # subr 4 performs hint replacement
  }
}

function tab2str(t, s) {
  for(i=1; i<=t[0]; ++i) s=s (s=="" ? "" : " ") t[i]
  return s
}

function mark_pfb_pos() {return ++pfb_text[0]}

function store_pfb(s) {pfb_text[++pfb_text[0]]=s}

function store_afm(s,n,m) {
  if (n==-1) afm_text_oth[++num_oth_chars] = s lig_data[m]
  else if (!(n in afm_text)) {
    ++num_enc_chars; afm_text[n]=s lig_data[m]; enc_name[n]=m
  } else mess("!CHAR " n " ALREADY EXISTS")
}

function get_data_files( s,key,a) {
  while (getline s < CD >0) {
    if (s~/^<\//) key=""
    if (key~/<PFB PROLOG>/)  PFB_PRO[++PFB_PRO[0]]=s
    if (key~/<PFB TRAILER>/) PFB_TRA[++PFB_TRA[0]]=s
    if (key~/<AFM PROLOG>/)  AFM_PRO[++AFM_PRO[0]]=s
    if (key~/<AFM TRAILER>/) AFM_TRA[++AFM_TRA[0]]=s
    if (key~/<AFM KPX>/)     AFM_KPX[++AFM_KPX[0]]=s
    if (s~/^<[^\/]/) key=s
  }
#
  while (getline s < FD >0) {
    if (split(s,a,/ +: +/)==2) font_data[a[1]]=a[2]
    else mess("!IMPROPER FONT INFO FILE")
  }
#
  while (getline s < KD >0) {
    if (s~/^KPX /) kpx_data[++kpx_data[0]]=s
    if (s~/^L /) {
      split(s,a); lig_data[a[2]]=lig_data[a[2]] " L " a[3] " " a[4] " ;"
    }
  }
}

function flush_pfb( s,i,n) {
  for (n=1; n<=PFB_PRO[0]; n++) {
    s=PFB_PRO[n]
    for (i in font_data)
      while (match(s,"##" i "##"))
        s=substr(s,1,RSTART-1) font_data[i] substr(s,RSTART+RLENGTH)
    if (match(s,"##OTHER_BLUES##")) s="" # no empty `other blues'
#
    if (match(s,"##ENCODING##")) {
      s=""
      for (i=0; i<=255; ++i) if (i in enc_name) {
        s=s (s==""?"":"\n") "dup " i "/" enc_name[i] " put"
      }
    }
    if (match(s,"##FONT_BOUNDING_BOX##"))
      s=substr(s,1,RSTART-1) mllx " " mlly " " murx " " mury\
        substr(s,RSTART+RLENGTH)
    if (match(s,"##STDHW##"))
      if (stdhw==0) s=""
      else s=substr(s,1,RSTART-1) stdhw substr(s,RSTART+RLENGTH)
    if (match(s,"##STDVW##"))
      if (stdvw==0) s=""
      else s=substr(s,1,RSTART-1) stdvw substr(s,RSTART+RLENGTH)
    if (match(s,"##STEMSNAPH##"))
      if (stdhw==0) s=""
      else s=substr(s,1,RSTART-1) tab2str(stemsnaph) substr(s,RSTART+RLENGTH)
    if (match(s,"##STEMSNAPV##"))
      if (stdvw==0) s=""
      else s=substr(s,1,RSTART-1) tab2str(stemsnapv) substr(s,RSTART+RLENGTH)
    if (match(s,"##NUMBER_OF_SUBRS##"))
      s=substr(s,1,RSTART-1) (1+subrs_base+xsubrs[0])\
        substr(s,RSTART+RLENGTH) # 1+subrs_base = number of standard subrs
    if (match(s,"##SUBROUTINES##")) {
      s=""
      for (i=1; i<=xsubrs[0]; ++i) {
        s=s (s=="" ? "" : "\n")\
          "dup " i+subrs_base " {\n" xsubrs[i] "\n\treturn\n\t} NP"
      }
    }
    if (match(s,"##NUMBER_OF_CHARSTRINGS##"))
      s=substr(s,1,RSTART-1) (num_enc_chars+num_oth_chars+1)\
        substr(s,RSTART+RLENGTH) # +1 because of /.notdef
    if (match(s,"##[^#]+##"))
      mess("!EXTRA TAG: " substr(s,RSTART,RLENGTH))
    if (s!="") print s > PFB
  }
#
  for (i=1; i<=pfb_text[0]; ++i)
    if (pfb_text[i]!="") # lines reserved for hints can be empty
      print pfb_text[i] > PFB
#
  for (n=1; n<=PFB_TRA[0]; n++) print PFB_TRA[n] > PFB
}

function flush_afm( s,i,a,n,
  font_dimen, dimen_name, header_byte, not_mathsy, not_mathex) {
  not_mathsy=not_mathex=1;
  for (n=0; n<=max_font_dimen; n++) for (i in font_data)
    if (match(i, "FONT_DIMEN" n "$")) {
      font_dimen[n]=font_data[i]; not_mathsy=(n>22); not_mathex=(n>13)
    } else if (n>=8) {not_mathsy=(n<=22); not_mathex=(n<=13)}
  prepare_fontdimen_names(!not_mathsy, !not_mathex)
  for (n=0; n<=max_font_dimen; n++) for (i in font_data)
    if (match(i, "DIMEN_NAME" n "$")) {
      dimen_name[n]=font_data[i]
      if (n in dimen_name_dft)
        sub(/\(default\)/, "(" dimen_name_dft[n] ")", dimen_name[n])
    }
  for (n=0; n<=max_header_byte; n++) {
    for (i in font_data)
      if (match(i, "HEADER_BYTE" n "$")) header_byte[n]=font_data[i]
  }
  for (n=1; n<=AFM_PRO[0]; n++) {
    s=AFM_PRO[n]
    if (match(s,"##FONT_BOUNDING_BOX##"))
      s=substr(s,1,RSTART-1) mllx " " mlly " " murx " " mury\
        substr(s,RSTART+RLENGTH)
    for (i in font_data)
      while (match(s,"##" i "##"))
        s=substr(s,1,RSTART-1) font_data[i] substr(s,RSTART+RLENGTH)
    if (match(s,"##PFM_NAME##")) s="" # no empty PFM data
    if (match(s,"TFM designsize")) {# append remaining TFM data:
      for (i=0; i<=max_font_dimen; ++i) if (i in font_dimen)
        s=s "\nComment TFM fontdimen " sprintf("%2d", i) ": "\
          sprintf("%-11s", font_dimen[i]) dimen_name[i]
      for (i=0; i<=max_header_byte; ++i) if (i in header_byte)
        s=s "\nComment TFM headerbyte " sprintf("%2d", i) ": " header_byte[i]
    }
    if (match(s,"##NUMBER_OF_CHARNAMES##"))
      s=substr(s,1,RSTART-1) (num_enc_chars+num_oth_chars)\
        substr(s,RSTART+RLENGTH)
    if (match(s,"##[^#]+##"))
      mess("!EXTRA TAG: " substr(s,RSTART,RLENGTH))
    if (s!="") print s > AFM
  }
#
  for (i=0; i<=255; ++i) if (i in afm_text) {
    print afm_text[i] > AFM
    delete afm_text[i]
  }
  for (i in afm_text) mess("!CHAR OF CODE " i)
  for (i=1; i<=num_oth_chars; ++i) print afm_text_oth[i] > AFM
#
  for (n=1; n<=AFM_TRA[0]; n++) {
    s=AFM_TRA[n]
    if (match(s,"##AFM_KPX##")) {
      s=""
      if (kpx_data[0]>0) flush_kpx(kpx_data)
    }
    if (s!="") print s > AFM
  }
}

function flush_kpx(k, s,i,a,l) {
  for (l=1; l<=AFM_KPX[0]; l++) {
    s=AFM_KPX[l]
    if (match(s,"##NUMBER_OF_KPX##"))
      s=substr(s,1,RSTART-1) k[0] substr(s,RSTART+RLENGTH)
    if (match(s,"##KPX_DATA##")) {
      s=""
      for(i=1; i<=k[0]; ++i) s=s (s==""?"":"\n") k[i]
    }
    print s > AFM
  }
}

function make_lineto(dx,dy) {
  if ((dx!=0) || (dy!=0)) {# emergency test
    if (dx==0) {store_pfb("\t" fround(dy) " vlineto")}
    else
      if (dy==0) {store_pfb("\t" fround(dx) " hlineto")}
      else {store_pfb("\t" fround(dx) " " fround(dy) " rlineto")}
  }
}

function make_moveto(dx,dy) {
  if ((dx!=0) || (dy!=0)) {# emergency test
    if (dx==0) {store_pfb("\t" fround(dy) " vmoveto")}
    else
      if (dy==0) {store_pfb("\t" fround(dx) " hmoveto")}
      else {store_pfb("\t" fround(dx) " " fround(dy) " rmoveto")}
  }
}

function make_curveto(dx1,dy1,dx2,dy2,dx3,dy3) {
  if ((dx1+dx2+dx3!=0) || (dy1+dy2+dy3!=0)) {# no loops (emergency test)
    if ((dy1==0) && (dx3==0))
      store_pfb("\t" fround(dx1) " " fround(dx2) " "\
        fround(dy2) " " fround(dy3) " hvcurveto")
    else
    if ((dx1==0) && (dy3==0))
      store_pfb("\t" fround(dy1) " " fround(dx2) " "\
      fround(dy2) " " fround(dx3) " vhcurveto")
    else
      store_pfb("\t" fround(dx1) " " fround(dy1) " " fround(dx2) " "\
        fround(dy2) " " fround(dx3) " " fround(dy3) " rrcurveto")
  }
}

function prepare_fontdimen_names(mathsy, mathex) {
# exactly one of the parameters equals 1, remaining one is equal to 0

  dimen_name_dft[0]="designsize"
  dimen_name_dft[1]="slant"
  dimen_name_dft[2]="space"
  dimen_name_dft[3]="space stretch"
  dimen_name_dft[4]="space shrink"
  dimen_name_dft[5]="xheight"
  dimen_name_dft[6]="quad"
  dimen_name_dft[7]="extra space"

  if (mathsy) {
    dimen_name_dft[8]="num1"
    dimen_name_dft[9]="num2"
    dimen_name_dft[10]="num3"
    dimen_name_dft[11]="denom1"
    dimen_name_dft[12]="denom2"
    dimen_name_dft[13]="sup1"
    dimen_name_dft[14]="sup2"
    dimen_name_dft[15]="sup3"
    dimen_name_dft[16]="sub1"
    dimen_name_dft[17]="sub2"
    dimen_name_dft[18]="supdrop"
    dimen_name_dft[19]="subdrop"
    dimen_name_dft[20]="delim1"
    dimen_name_dft[21]="delim2"
    dimen_name_dft[22]="axis height"
  }

  if (mathex) {
    dimen_name_dft[8]="default rule thickness"
    dimen_name_dft[9]="bigop spacing1"
    dimen_name_dft[10]="bigop spacing2"
    dimen_name_dft[11]="bigop spacing3"
    dimen_name_dft[12]="bigop spacing4"
    dimen_name_dft[13]="bigop spacing5"
  }

}

function rational(x,nlimit,dlimit, p0,q0,p1,q1,p2,q2,s,ds) {
  # a simplified version of the code kindly sent us by Berthold K. P. Horn
  if (x == 0.0) return "0 1" # if (x == 0.0) return "0/1"
  if (x < 0.0) return "-" rational(-x, nlimit, dlimit)
  # if (int(x) > nlimit) return round(x) # unlikely (in Type 1 fonts)
  p0=0; q0=1; p1=1; q1=0; s=x;
  while (1) {
    # in general, it is advisable to avoid crossing limits;
    # here, it is perhaps a bit of pedantry:
    if ( int(s)!=0 &&\
      ((p1 > (nlimit-p0)/int(s)) || (q1 > (dlimit-q0)/int(s))) )
      {p2=p1; q2=q1; break}
    p2=p0+int(s)*p1; q2=q0+int(s)*q1;
#   printf("%ld %ld %ld %ld %ld %ld %lg\n", p0, q0, p1, q1, p2, q2, s)
    if (p2/q2==x) break
    ds=s-int(s)
    if (ds == 0.0) break
    p0=p1; q0=q1; p1=p2; q1=q2; s=1/ds
  }
  # q2 != 0
  return p2 " " q2 # the answer is p2/q2
}

function approx(x, r,a) {
  # 32767 is a limit due to old (DOS?) implementations of t1utils
  r=rational(x,32767,32767); split(r,a); return (a[2]==1 ? a[1] : r " div")
}

function abs(x) {return (x>=0 ? x : -x)}

function round(x) {# round
  return (x>=0 ? int(x+.5) : -int(-x+.5))
}

function fround(x) {# forced round
  if (x!=int(x)) {
    mess("!WRONG ROUNDING IN METAPOST " x " IN " name)
    return (x>=0 ? int(x+.5) : -int(-x+.5))
  } else return x
}

function mess(s) {print s > CONSOLE; if (LOG!="") print s > LOG}

